"""
Debug page.
"""
import urwid

from clay.pages.page import AbstractPage
from clay.log import Logger
from clay.clipboard import copy
from clay.gp import GP


class DebugItem(urwid.AttrMap):
    """
    Represents a single debug log item.
    """
    def selectable(self):
        return True

    def __init__(self, log_record):
        self.log_record = log_record

        self.columns = urwid.Columns([
            ('pack', urwid.Text(self.log_record.verbosity.ljust(8))),
            urwid.Text(
                (
                    self.log_record.formatted_timestamp +
                    '\n' +
                    self.log_record.formatted_message
                )
            )
        ])

        super(DebugItem, self).__init__(self.columns, 'panel', 'panel_focus')

    def keypress(self, _, key):
        """
        Handle heypress.
        """
        if key == 'enter':
            copy(self.log_record.formatted_message)
            return None
        return key


class DebugPage(urwid.Pile, AbstractPage):
    """
    Represents debug page.
    """
    def __init__(self, app):
        self.app = app
        self.walker = urwid.SimpleListWalker([])
        for log_record in Logger.get().get_logs():
            self._append_log(log_record)
        Logger.get().on_log_event += self._append_log
        self.listbox = urwid.ListBox(self.walker)

        self.debug_data = urwid.Text('')

        super(DebugPage, self).__init__([
            ('pack', self.debug_data),
            ('pack', urwid.Text('')),
            ('pack', urwid.Text('Hit "Enter" to copy selected message to clipboard.')),
            ('pack', urwid.Divider(u'\u2550')),
            self.listbox
        ])

        GP.get().auth_state_changed += self.update

        self.update()

    def update(self, *_):
        """
        Update this widget.
        """
        gpclient = GP.get()
        self.debug_data.set_text(
            '- Is authenticated: {}\n'
            '- Is subscribed: {}'.format(
                gpclient.is_authenticated,
                gpclient.is_subscribed if gpclient.is_authenticated else None
            )
        )

    def _append_log(self, log_record):
        """
        Add log record to list.
        """
        self.walker.insert(0, urwid.Divider(u'\u2500'))
        self.walker.insert(0, DebugItem(log_record))

    @property
    def name(self):
        """
        Return page name.
        """
        return "Debug"

    @property
    def key(self):
        """
        Return page key (``int``), used for hotkeys.
        """
        return 0

    def activate(self):
        """
        Notify page that it is activated.
        """
        pass



"""
Clipboard utils.
"""
from subprocess import Popen, PIPE

from clay.notifications import NotificationArea


COMMANDS = [
    ('xclip', '-selection', 'clipboard'),
    ('xsel', '-bi'),
]


def copy(text):
    """
    Copy text to clipboard.

    Return True on success.
    """
    for cmd in COMMANDS:
        proc = Popen(cmd, stdin=PIPE)
        proc.communicate(text.encode('utf-8'))
        if proc.returncode == 0:
            return True

    NotificationArea.notify(
        'Failed to copy text to clipboard. '
        'Please install "xclip" or "xsel".'
    )
    return False



"""
Predefined values.
"""
APP_NAME = 'Clay Player'
VERSION = '0.7.1'
AUTHOR = "Andrew Dunai"
DESCRIPTION = "Awesome standalone command line player for Google Play Music"

try:
    from codename import codename
except ImportError:
    VERSION_WITH_CODENAME = VERSION
else:
    VERSION_WITH_CODENAME = VERSION + ' (' + codename(separator=' ', id=VERSION) + ')'

COPYRIGHT_MESSAGE = """{app_name} {version}

Copyright 2017 - {year} (C) {author} and {app_name} contributors.
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licences/gpl.html>
This is free software; you are free to change it and redistribute it.
There is NO WARRANTY, to the extent permitted by law
""".format(app_name=APP_NAME, year=2018,
           version=VERSION_WITH_CODENAME, author=AUTHOR)

USER_AGENT = ' '.join([
    'Mozilla/5.0 (X11; Linux x86_64)'
    'AppleWebKit/537.36 (KHTML, like Gecko)'
    'Chrome/62.0.3202.94 Safari/537.36'
])



"""
Logger implementation.
"""
# pylint: disable=too-few-public-methods
from threading import Lock
from datetime import datetime

from clay.eventhook import EventHook


class LoggerRecord(object):
    """
    Represents a logger record.
    """
    def __init__(self, verbosity, message, args):
        self._timestamp = datetime.now()
        self._verbosity = verbosity
        self._message = message
        self._args = args

    @property
    def formatted_timestamp(self):
        """
        Return timestamp.
        """
        return str(self._timestamp)

    @property
    def verbosity(self):
        """
        Return verbosity.
        """
        return self._verbosity

    @property
    def formatted_message(self):
        """
        Return formatted message.
        """
        return self._message % self._args


class Logger(object):
    """
    Global logger.

    Allows subscribing to log events.
    """
    instance = None

    def __init__(self):
        assert self.__class__.instance is None, 'Can be created only once!'

        self.logs = []
        self.logfile = open('/tmp/clay.log', 'w')

        self._lock = Lock()

        self.on_log_event = EventHook()

    @classmethod
    def get(cls):
        """
        Create new :class:`.Logger` instance or return existing one.
        """
        if cls.instance is None:
            cls.instance = Logger()

        return cls.instance

    def log(self, level, message, *args):
        """
        Add log item.
        """
        self._lock.acquire()
        try:
            logger_record = LoggerRecord(level, message, args)
            self.logs.append(logger_record)
            self.logfile.write('{} {:8} {}\n'.format(
                logger_record.formatted_timestamp,
                logger_record.verbosity,
                logger_record.formatted_message
            ))
            self.logfile.flush()
            self.on_log_event.fire(logger_record)
        finally:
            self._lock.release()

    def debug(self, message, *args):
        """
        Add debug log item.
        """
        self.log('DEBUG', message, *args)

    def info(self, message, *args):
        """
        Add info log item.
        """
        self.log('INFO', message, *args)

    def warn(self, message, *args):
        """
        Add warning log item.
        """
        self.log('WARNING', message, *args)

    warning = warn

    def error(self, message, *args):
        """
        Add error log item.
        """
        self.log('ERROR', message, *args)

    def get_logs(self):
        """
        Return all logs.
        """
        return self.logs
